@@ -1,18 +1,18 @@
 import re
 import enum
 import json
+import types
 import shutil
+import typing
 import pathlib
-import argparse
 import tempfile
+import itertools
 import collections
 import unicodedata
 
 from tabulate import tabulate
 import segments
 import attr
-from clldutils.misc import nfilter
-from clldutils.lgr import ABBRS
 from csvw.dsv import UnicodeWriter, reader
 from csvw.metadata import Link
 from pycldf import Dataset
@@ -25,10 +25,10 @@ except ImportError:  # pragma: no cover
 
 from pyigt.util import expand_standard_abbr
 from pyigt.lgrmorphemes import (
-    GlossedWord, split_morphemes, MORPHEME_SEPARATORS, remove_morpheme_separators,
+    GlossedWord, split_morphemes, remove_morpheme_separators, GlossedMorpheme
 )
 
-__all__ = ['IGT', 'Corpus', 'CorpusSpec', 'LGRConformance']
+__all__ = ['IGT', 'Corpus', 'LGRConformance']
 
 NON_OVERT_ELEMENT = '∅'
 
@@ -42,109 +42,20 @@ def with_lingpy():
 
 @enum.unique
 class LGRConformance(enum.IntEnum):
+    """
+    Conformance levels with respect to alignment of phrase and gloss of an `IGT`.
+
+    We distinguish the following levels:
+
+    - morpheme-aligned (IGT conforms to LGR Rule 2)
+    - word-aligned (IGT conforms to LGR Rule 1, but not Rule 2)
+    - unaligned (IGT does not conform to LGR Rule 1)
+    """
     MORPHEME_ALIGNED = 2
     WORD_ALIGNED = 1
     UNALIGNED = 0
 
 
-def _morpheme_and_infixes(m):
-    morpheme, infixes, infix, in_infix = [], [], [], False
-
-    if m.startswith('<'):  # pragma: no cover
-        morpheme.append('')
-
-    for c in m:
-        if c == '<':
-            in_infix = True
-            continue
-        elif c == '>' and in_infix:
-            infixes.append(''.join(infix))
-            infix = []
-            in_infix = False
-            continue
-
-        if in_infix:
-            infix.append(c)
-        else:
-            morpheme.append(c)
-
-    if in_infix:
-        return [m]
-    return [''.join(morpheme)] + infixes
-
-
-def iter_morphemes(s, split_infixes=True):
-    """
-    Split word into morphemes following the Leipzig Glossing Rules.
-    """
-    morpheme, separator = [], None
-
-    for c in (s or ''):
-        if c in MORPHEME_SEPARATORS:
-            if split_infixes:
-                yield from _morpheme_and_infixes(''.join(morpheme))
-            else:
-                yield (separator, ''.join(morpheme))
-            morpheme = []
-            separator = c
-        else:
-            morpheme.append(c)
-
-    if split_infixes:
-        yield from _morpheme_and_infixes(''.join(morpheme))
-    else:
-        yield (separator, ''.join(morpheme))
-
-
-@attr.s
-class CorpusSpec(object):
-    punctuation = attr.ib(
-        validator=attr.validators.instance_of(list),
-        default=list(',;.”“"()'))
-    concept_replace = attr.ib(
-        validator=attr.validators.instance_of(dict),
-        default={'.': ' ', '†(': '', '†': ''})
-    paradigm_marker = attr.ib(
-        validator=attr.validators.instance_of(str),
-        default=':')
-    # A "simple" `CorpusSpec`, only recognizing one single-character morpheme separator, can be
-    # constructed by setting `morpheme_separator`.
-    morpheme_separator = attr.ib(
-        validator=attr.validators.optional(attr.validators.instance_of(str)),
-        default=None)
-    label_pattern = attr.ib(default=re.compile('^([A-Z][A-Z0-9]*|([1-3](DL|PL|SG|DU)))$'))
-
-    def split_morphemes(self, s, split_infixes=True):
-        if self.morpheme_separator:
-            return s.split(self.morpheme_separator)
-        return list(iter_morphemes(s, split_infixes=split_infixes))
-
-    def strip_punctuation(self, s):
-        for p in self.punctuation:
-            s = s.replace(p, '')
-        return s
-
-    def is_grammatical_gloss_label(self, s):
-        return bool((s in ABBRS) or self.label_pattern.match(s))
-
-    def clean_concept(self, c):
-        for k, v in self.concept_replace.items():
-            c = c.replace(k, v)
-        return c.strip()
-
-    def _glosses(self, concept, ctype):
-        return [
-            g for g in nfilter(self.clean_concept(cn) for cn in concept.split(self.paradigm_marker))
-            if (ctype == 'grammar' and self.is_grammatical_gloss_label(g))
-            or (ctype == 'lexicon' and not self.is_grammatical_gloss_label(g))]  # noqa: W503
-
-    def lexical_gloss(self, concept):
-        return ' // '.join(self._glosses(concept, 'lexicon'))
-
-    def grammatical_glosses(self, concept):
-        return self._glosses(concept, 'grammar')
-
-
 def parse_phrase(p):
     """
     We must take LGR Rule 2A into account, i.e. attach morphemes separated by " -" to the
@@ -160,11 +71,57 @@ def parse_phrase(p):
 @attr.s
 class IGT(object):
     """
-    Interlinear Glossed Text
-
     The main trait of IGT is the alignment of words and glosses. Thus, we are mostly interested
     in the two aligned "lines": the analyzed text and the glosses, rather than trying to support
     any number of tiers, and alignment based on timestamps or similar.
+    Thus, an `IGT` instance is a `list` of aligned words, and each aligned word a `list` of aligned
+    morphemes. This structure can be exploited to access parts of the alignment, see
+    :meth:`IGT.__getitem__`
+
+    :ivar phrase: `list` of `str` representing the gloss-aligned words of the IGT.
+    :ivar gloss: `list` of `str` representing the word-aligned glosses of the IGT.
+    :ivar id: Optional identifier, can be used for referencing the `IGT` if it part of a `Corpus`.
+    :ivar properties: `typing.Dict[str, object]` storing additional properties of an `IGT`, e.g. \
+    additional column values read from a row in a CLDF ExampleTable.
+    :ivar language: Optional language identifier, specifying the object language of the `IGT`.
+    :ivar translation: Optional translation of the phrase.
+    :ivar abbrs: Optional `dict` providing descriptions of gloss labels used in the `IGT`.
+    :ivar strict: `bool` flag signaling whether to parse the `IGT` in strict mode, i.e. requiring \
+    matching morpheme separators in phrase and gloss, or not.
+
+    .. note::
+
+        **LGR Conformance**
+
+        While the main purpose of an `IGT` is providing access to its words, morphemes and glosses,
+        it also supports error/conformance checking. Thus, it is possible to initialize an `IGT`
+        with "broken" data.
+
+        .. code-block:: python
+
+            >>> from pyigt import IGT
+            >>> igt = IGT(phrase='two words', gloss='ONE.GLOSS')
+            >>> igt.conformance
+            <LGRConformance.UNALIGNED: 0>
+
+        So before processing `IGT` instances, it should be checked whether the conformance level
+        (see :class:`LGRConformance`) of the `IGT` is sufficient for the downstream requirements.
+        Otherwise, accessing properties like :meth:`IGT.glossed_words` may lead to unexpected
+        results:
+
+        .. code-block:: python
+
+            >>> igt.glossed_words  # we extract as many glossed words as possible ...
+            [<GlossedWord word=a gloss=C>]
+            >>> len(igt)
+            1
+            >>> len(igt.phrase)
+            2
+            >>> igt = IGT(phrase='multi-morph', gloss='GLOSS')
+            >>> igt.conformance
+            <LGRConformance.WORD_ALIGNED: 1>
+            >>> igt[0].glossed_morphemes  # we extract as many glossed morphemes as possible ...
+            [<GlossedMorpheme morpheme=multi gloss=GLOSS>]
     """
     phrase = attr.ib(
         validator=attr.validators.instance_of(list),
@@ -178,7 +135,6 @@ class IGT(object):
     properties = attr.ib(validator=attr.validators.instance_of(dict), default=attr.Factory(dict))
     language = attr.ib(default=None)
     translation = attr.ib(default=None)
-    spec = attr.ib(default=CorpusSpec())
     abbrs = attr.ib(validator=attr.validators.instance_of(dict), default=attr.Factory(dict))
     strict = attr.ib(default=False)
 
@@ -199,15 +155,26 @@ class IGT(object):
                 # Punctuation, Final quote
                 self.translation = self.translation[:-1].strip()
 
+    def __len__(self):
+        return len(self.glossed_words)
+
+    def __iter__(self):
+        yield from self.glossed_words
+
     @property
-    def glossed_words(self):
+    def glossed_words(self) -> typing.List[GlossedWord]:
         return [GlossedWord(w, g, strict=self.strict) for w, g in zip(self.phrase, self.gloss)]
 
     @property
-    def prosodic_words(self):
+    def prosodic_words(self) -> typing.List[GlossedWord]:
         """
-        1. Split prosodically free elements marked with " -" separator.
-        2. Conflate clitics.
+        Interpret an IGT's phrase prosodically, i.e.
+
+        1. splits prosodically free elements marked with " -" separator and
+        2. conflates clitics.
+
+        Use :meth:`IGT.as_prosodic` to get an `IGT` instance initialised from the prosodic words
+        of an `IGT` instance.
         """
         res = []
         for w, g in zip(self.phrase, self.gloss):
@@ -227,10 +194,15 @@ class IGT(object):
         return res
 
     @property
-    def morphosyntactic_words(self):
+    def morphosyntactic_words(self) -> typing.List[GlossedWord]:
         """
-        1. Conflate prosodically free elements marked with " -" separator.
-        2. Split clitics into separate words.
+        Interpret an IGT's phrase morphosyntactically, i.e.
+
+        1. conflate prosodically free elements marked with " -" separator and
+        2. split clitics into separate words.
+
+        Use :meth:`IGT.as_morphosyntactic` to get an `IGT` instance initialised from the
+        morphosyntactic words of an `IGT` instance.
         """
         res = []
         for w, g in zip(self.phrase, self.gloss):
@@ -239,14 +211,62 @@ class IGT(object):
                 for ww, gg in zip(w.split('='), g.split('='))])
         return res
 
+    def as_prosodic(self) -> 'IGT':
+        """
+        .. code-block:: python
+
+            >>> from pyigt import IGT
+            >>> igt = IGT(phrase='a=bcd -e', gloss='A=BCD-E')
+            >>> len(igt) != len(igt.as_prosodic())
+            True
+            >>> igt[0].word
+            'a=bcd -e'
+            >>> igt.as_prosodic()[0].word
+            'a=bcd'
+        """
+        return IGT(
+            phrase=[gw.word for gw in self.prosodic_words],
+            gloss=[gw.gloss for gw in self.prosodic_words],
+            id=self.id,
+            properties=self.properties,
+            language=self.language,
+            translation=self.translation,
+            abbrs=self.abbrs,
+            strict=self.strict,
+        )
+
+    def as_morphosyntactic(self):
+        """
+        .. code-block:: python
+
+            >>> from pyigt import IGT
+            >>> igt = IGT(phrase='a=bcd -e', gloss='A=BCD-E')
+            >>> len(igt) != len(igt.as_morphosyntactic())
+            True
+            >>> igt[0].word
+            'a=bcd -e'
+            >>> igt.as_morphosyntactic()[-1].word
+            'bcd -e'
+        """
+        return IGT(
+            phrase=[gw.word for gw in self.morphosyntactic_words],
+            gloss=[gw.gloss for gw in self.morphosyntactic_words],
+            id=self.id,
+            properties=self.properties,
+            language=self.language,
+            translation=self.translation,
+            abbrs=self.abbrs,
+            strict=self.strict,
+        )
+
     @property
-    def gloss_abbrs(self):
+    def gloss_abbrs(self) -> collections.OrderedDict:
         res = collections.OrderedDict()
         for gw in self.glossed_words:
             for gm in gw:
                 for element in gm.gloss.elements:
                     # We disregard "I".
-                    if element != 'I' and self.spec.is_grammatical_gloss_label(str(element)):
+                    if element != 'I' and element.is_category_label:
                         if element in self.abbrs:
                             res[element] = self.abbrs[element]
                         else:
@@ -255,6 +275,10 @@ class IGT(object):
         return res
 
     def __str__(self):
+        """
+        A plain text representation of the IGT, to be viewed with a monospaced font to make
+        alignments work.
+        """
         return '{0}\n{1}{2}'.format(
             self.primary_text,
             tabulate([self.gloss], self.phrase, tablefmt='plain'),
@@ -268,36 +292,56 @@ class IGT(object):
             abbrs = ''.join('\n  {} = {}'.format(k.ljust(mlen), v) for k, v in abbrs)
         print('{}{}'.format(self, abbrs or ''))
 
-    def __getitem__(self, item):
+    def __getitem__(self, i: typing.Union[int, typing.Tuple[int, typing.Union[int, slice]]]) \
+            -> typing.Union[typing.List, GlossedWord, GlossedMorpheme]:
         """
-        Provide access to individual (word, gloss) or (morpheme, gloss) pairs.
+        Provide access to `GlossedWord` or `GlossedMorpheme` (s) by zero-based index.
 
-        :param item: An `int` index to reference a (word, gloss) pair or a (`int`, `int`) tuple,\
-        referencing a (morpheme, gloss) pair.
-        :return: (word, gloss) or (morpheme, gloss) pair
+        :param i: An `int` index to reference a `GlossedWord` or a (`int`, `int`) tuple,\
+        referencing a `GlossedMorpheme`.
+
+        .. code-block:: python
+
+            >>> from pyigt import IGT
+            >>> igt = IGT(phrase="zəp-le: ȵi-ke: pe-ji qeʴlotʂu-ʁɑ,", gloss="a-DEF b-IN c-CSM d-LO")
+            >>> igt[0].word
+            'zəp-le:'
+            >>> [gw.word for gw in igt[2:]]
+            ['pe-ji', 'qeʴlotʂu-ʁɑ,']
+            >>> str(igt[0, 0].morpheme)
+            'zəp'
+            >>> [str(gm.morpheme) for gm in igt[1, 0:]]  # All morphemes of the second word
+            ['ȵi', 'ke:']
+            >>> [str(gm.morpheme) for gm in igt[0:, 0]]  # First morpheme in each word
+            ['zəp', 'ȵi', 'pe', 'qeʴlotʂu']
         """
-        if not isinstance(item, tuple):
-            gw = self.glossed_words[item]
-            return gw.word, gw.gloss
-        gm = self.glossed_words[item[0]][item[1]]
-        return gm.morpheme, gm.gloss
+        if isinstance(i, tuple):
+            assert len(i) == 2
+            word = self.glossed_words[i[0]]
+            if isinstance(word, list):
+                return [w[i[1]] for w in word]
+            return word[i[1]]
+        return self.glossed_words[i]
 
     @property
-    def conformance(self):
+    def conformance(self) -> LGRConformance:
+        """
+        Alignment level of the `IGT`.
+        """
         if self.is_valid(strict=True):
             return LGRConformance.MORPHEME_ALIGNED
         if self.is_valid():
             return LGRConformance.WORD_ALIGNED
         return LGRConformance.UNALIGNED
 
-    def is_valid(self, strict=False):
+    def is_valid(self, strict: bool = False) -> bool:
         try:
             self.check(strict=strict)
             return True
         except (ValueError, AssertionError):
             return False
 
-    def check(self, strict=False, verbose=False):
+    def check(self, strict: bool = False, verbose: bool = False):
         """
         :param strict: If `True`, also check Rule 2: Morpheme-by-morpheme correspondence.
         """
@@ -321,11 +365,14 @@ class IGT(object):
                         'glosses!')
 
     @property
-    def phrase_text(self):
+    def phrase_text(self) -> str:
         return ' '.join([w or '' for w in self.phrase])
 
     @property
-    def primary_text(self):
+    def primary_text(self) -> str:
+        """
+        The primary text of the `IGT`, i.e. the phrase stripped off morpheme separators.
+        """
         try:
             words = []
             for gw in self.glossed_words:
@@ -335,24 +382,29 @@ class IGT(object):
             return remove_morpheme_separators(self.phrase_text)
 
     @property
-    def gloss_text(self):
+    def gloss_text(self) -> str:
         return ' '.join(self.gloss)
 
 
-class Concordances(argparse.Namespace):
-    def __getitem__(self, item):
-        return vars(self)[item]
+def _clean_lexical_concept(s):
+    s = re.sub(r'†\(([^)]+)\)', lambda m: m.groups()[0], s)
+    return s.replace('†', '').strip()
 
 
 class Corpus(object):
     """
     A Corpus is an immutable, ordered list of `IGT` instances.
+
+    It provides access to concordance-like aggregated statistics of its texts.
+
+    :ivar monolingual: Flag signaling whether the corpus is monolingual or contains `IGT` from \
+    different object languages.
     """
-    def __init__(self, igts, spec=None, fname=None):
+    def __init__(self, igts: typing.Iterable[IGT], fname=None, clean_lexical_concept=None):
+        self.clean_lexical_concept = clean_lexical_concept or _clean_lexical_concept
         self.fname = fname
-        self.spec = spec or CorpusSpec()
-        self._igts = collections.OrderedDict([(igt.id, igt) for igt in igts])
-        self._concordances = Concordances(
+        self._igts = collections.OrderedDict([(igt.id or n, igt) for n, igt in enumerate(igts)])
+        self._concordance = dict(
             grammar=collections.defaultdict(list),
             lexicon=collections.defaultdict(list),
             form=collections.defaultdict(list),
@@ -360,27 +412,71 @@ class Corpus(object):
         # Since changing the IGTs in the corpus is not allowed, we can compute concordances right
         # away.
         for idx, igt in self._igts.items():
-            if not igt.is_valid(strict=True):
+            if not igt.is_valid(strict=True):  # We ignore non-morpheme-aligned IGTs.
                 continue
-            for i, gw in enumerate(igt.glossed_words):
-                if not gw.is_valid:
-                    continue  # pragma: no cover
+            for i, gw in enumerate(igt):
                 for j, gm in enumerate(gw):
-                    morpheme, concept = str(gm.morpheme), str(gm.gloss)
-                    morpheme = self.spec.strip_punctuation(morpheme)
-                    if not morpheme:
+                    if not gm.form:
                         continue
+
                     ref = (idx, i, j)
-                    for g in self.spec.grammatical_glosses(concept):
-                        self._concordances.grammar[morpheme, g, concept, igt.language].append(ref)
-                    self._concordances.lexicon[
-                        morpheme, self.spec.lexical_gloss(concept), concept, igt.language
-                    ].append(ref)
-                    self._concordances.form[morpheme, concept, concept, igt.language].append(ref)
+                    for g in gm.grammatical_concepts:
+                        self._concordance['grammar'][g].append(ref)
+                    self._concordance['lexicon'][' // '.join(gm.lexical_concepts)].append(ref)
+                    self._concordance['form'][gm.form].append(ref)
         self.monolingual = len(set(igt.language for igt in self._igts.values())) == 1
 
+    @property
+    def grammar(self) -> typing.Dict[str, typing.List[typing.Tuple[int, int, int]]]:
+        """
+        Maps grammatical concepts to lists of occurrences.
+
+        .. code-block:: python
+
+            >>> from pyigt import Corpus, IGT
+            >>> igt = IGT(phrase="ni-c-chihui-lia in no-piltzin ce calli",
+            ...           gloss="1SG.SUBJ-3SG.OBJ-mach-APPL DET 1SG.POSS-Sohn ein Haus")
+            >>> c = Corpus([igt])
+            >>> [[c[ref] for ref in c.grammar[k]] for k in c.grammar if k.startswith('1SG')]
+            [[<GlossedMorpheme morpheme=ni gloss=1SG.SUBJ>],
+             [<GlossedMorpheme morpheme=no gloss=1SG.POSS>]]
+        """
+        return self._concordance['grammar']
+
+    @property
+    def lexicon(self) -> typing.Dict[str, typing.List[typing.Tuple[int, int, int]]]:
+        """
+        Maps lexical concepts to lists of occurrences.
+
+        .. code-block:: python
+
+            >>> from pyigt import Corpus, IGT
+            >>> igt = IGT(phrase="ni-c-chihui-lia in no-piltzin ce calli",
+            ...           gloss="1SG.SUBJ-3SG.OBJ-mach-APPL DET 1SG.POSS-Sohn ein Haus")
+            >>> c = Corpus([igt])
+            >>> [c[ref] for ref in c.lexicon['Sohn']]
+            [<GlossedMorpheme morpheme=piltzin gloss=Sohn>]
+        """
+        return self._concordance['lexicon']
+
+    @property
+    def form(self) -> typing.Dict[str, typing.List[typing.Tuple[int, int, int]]]:
+        """
+        Maps grammatical concepts to lists of occurrences.
+
+        .. code-block:: python
+
+            >>> from pyigt import Corpus, IGT
+            >>> igt = IGT(phrase="ni-c-chihui-lia in no-piltzin ce calli",
+            ...           gloss="1SG.SUBJ-3SG.OBJ-mach-APPL DET 1SG.POSS-Sohn ein Haus")
+            >>> c = Corpus([igt])
+            >>> [k for k in c.form]
+            ['ni', 'c', 'chihui', 'lia', 'in', 'no', 'piltzin', 'ce', 'calli']
+        """
+        return self._concordance['form']
+
     @staticmethod
-    def get_column_names(cldf):
+    def get_column_names(cldf: Dataset) -> types.SimpleNamespace:
         # We lookup local column names by ontology term:
         lookup = [
             ('id', 'id'),
@@ -389,20 +485,18 @@ class Corpus(object):
             ('translation', 'translatedText'),
             ('language', 'languageReference'),
         ]
-        names = collections.namedtuple('colnames', [k for k, v in lookup])
-        return names(**{
+        return types.SimpleNamespace(**{
             k: cldf['ExampleTable', v].name if ('ExampleTable', v) in cldf else None
             for k, v in lookup})
 
     @classmethod
-    def from_cldf(cls, cldf, spec=None):
+    def from_cldf(cls, cldf: Dataset) -> 'Corpus':
         """
-        A corpus of IGT examples provided with a CLDF dataset.
+        Instantiate a corpus of IGT examples from a CLDF dataset.
 
         :param cldf: a `pycldf.Dataset` instance.
         :param spec: a `CorpusSpec` instance, specifying how to interpret markup in the corpus.
         """
-        spec = spec or CorpusSpec()
         cols = cls.get_column_names(cldf)
         igts = [
             IGT(
@@ -412,21 +506,18 @@ class Corpus(object):
                 language=igt.get(cols.language),
                 translation=igt.get(cols.translation),
                 properties=igt,
-                spec=spec,
             )
             for igt in cldf['ExampleTable']]
         return cls(
             igts,
-            spec=spec,
             fname=cldf.tablegroup._fname.parent / str(cldf['ExampleTable'].url))
 
     @classmethod
-    def from_stream(cls, stream, spec=None):
+    def from_stream(cls, stream) -> 'Corpus':
         from csvw.metadata import TableGroup
         cldf = Dataset(TableGroup(fname=pathlib.Path('tmp.json')))
         cldf.add_component('ExampleTable')
 
-        spec = spec or CorpusSpec()
         cols = cls.get_column_names(cldf)
         igts = [
             IGT(
@@ -435,25 +526,23 @@ class Corpus(object):
                 phrase=igt[cols.phrase].split('\\t'),
                 language=igt.get(cols.language),
                 properties=igt,
-                spec=spec,
             )
             for igt in reader(stream.read().splitlines(), dicts=True)]
-        return cls(igts, spec=spec)
+        return cls(igts)
 
     @classmethod
-    def from_path(cls, path, spec=None):
+    def from_path(cls, path: typing.Union[str, pathlib.Path]) -> 'Corpus':
         """
         Instantiate a corpus from a file path.
 
         :param path: Either a path to a CLDF dataset's metadata file or to a CLDF Examples \
         component as CSV file. Note that in the latter case, the file must use the default \
         column names, as defined in the CLDF ontology.
-        :return: `Corpus` instance.
         """
         if isinstance(path, str):
             path = pathlib.Path(path)
         if path.suffix == '.json':
-            return cls.from_cldf(Dataset.from_metadata(path), spec=spec)
+            return cls.from_cldf(Dataset.from_metadata(path))
         # We are given only an ExampleTable. Let's create the appropriate dataset:
         header = None
         for d in reader(path, dicts=True):
@@ -467,7 +556,7 @@ class Corpus(object):
         default_cols = [col.name for col in t.tableSchema.columns]
         ds.remove_columns(t, *list(set(default_cols) - set(header)))
         ds.add_columns(t, *list(set(header) - set(default_cols)))
-        return cls.from_cldf(ds, spec=spec)
+        return cls.from_cldf(ds)
 
     def __len__(self):
         return len(self._igts)
@@ -477,38 +566,53 @@ class Corpus(object):
 
     def __getitem__(self, item):
         if not isinstance(item, tuple):
-            return self._igts[item]
+            return self._igts[item] if item in self._igts else list(self._igts.values())[item]
         if len(item) == 2:
             return self._igts[item[0]][item[1]]
-        return self._igts[item[0]][tuple(item[1:])]
+        return self[item[0]][tuple(item[1:])]
 
     def get_stats(self):
-        wordc, morpc = 0, 0
-        for igt in self:
-            for word in igt.phrase:
-                wordc += 1
-                morpc += len(self.spec.split_morphemes(word))
-        return len(self._igts), wordc, morpc
+        return (
+            len(self),
+            sum(len(igt) for igt in self),
+            sum(sum(len(w) for w in igt) for igt in self))
 
     def get_lgr_conformance_stats(self):
         return collections.Counter([igt.conformance for igt in self])
 
-    def write_concordance(self, ctype, filename=None):
+    def write_concordance(self, ctype: str, filename=None):
+        """
+        :param ctype: `lexicon` or `grammar` or `form`.
+        """
+        conc = collections.defaultdict(list)
+        for c, refs in getattr(self, ctype).items():
+            for ref in refs:
+                # We want one row per unique (form, language, concept, gloss).
+                if ctype == 'form':
+                    gloss = str(self[ref].gloss)
+                    conc[c, gloss, gloss, self[ref[0]].language].append(ref)
+                else:
+                    conc[
+                        self[ref].form,
+                        self.clean_lexical_concept(c),
+                        c,
+                        self[ref[0]].language].append(ref)
+
         with UnicodeWriter(filename, delimiter='\t') as w:
             h = ['ID', 'FORM', 'GLOSS', 'GLOSS_IN_SOURCE', 'OCCURRENCE', 'REF']
             if not self.monolingual:
                 h.insert(1, 'LANGUAGE_ID')
             w.writerow(h)
-
-            for i, (k, v) in enumerate(sorted(
-                    self._concordances[ctype].items(), key=lambda x: (-len(x[1]), x[0])), start=1):
+            # We order the rows by descending frequency:
+            for i, (k, refs) in enumerate(
+                    sorted(conc.items(), key=lambda x: (-len(x[1]), x[0])), start=1):
                 c = [
                     i,
                     k[0],
                     k[1],
                     k[2],
-                    len(v),
-                    ' '.join(['{0}:{1}:{2}'.format(x, y, z) for x, y, z in v])]
+                    len(refs),
+                    ' '.join(['{}:{}:{}'.format(*ref) for ref in refs])]
                 if not self.monolingual:
                     c.insert(1, k[3])
                 w.writerow(c)
@@ -516,46 +620,32 @@ class Corpus(object):
         if not filename:
             print(w.read().decode('utf8'))
 
-    def get_concepts(self, ctype='lexicon'):
-        concepts = collections.defaultdict(list)
-
-        for (form, c1, c2, l), occs in self._concordances[ctype].items():
-            # get occurrence, and one example
-            assert form
-            concepts[c1].append((form, l, c2, len(occs)))
-
-        return concepts
-
     def write_concepts(self, ctype, filename=None):
-        def format_form(f):
-            if self.monolingual:
-                return f[0]
-            return '{}: {}'.format(f[1], f[0])
+        """
+        :param ctype: `lexicon` or `grammar`.
+        """
+        def form(ref):
+            return self[ref].form if self.monolingual else '{}: {}'.format(
+                self[ref[0]].language, self[ref].form)
 
-        concepts = self.get_concepts(ctype)
-        with UnicodeWriter(filename, delimiter='\t') as w:
-            w.writerow(
-                ['ID', 'ENGLISH', 'OCCURRENCE', 'CONCEPT_IN_SOURCE', 'FORMS', 'PHRASE', 'GLOSS'])
-            for i, (concept, forms) in enumerate(
-                    sorted(concepts.items(), key=lambda x: x[1][0][-1], reverse=True), start=1):
-                # Get the IGT containing the first occurrence listed in the concordance as example:
-                igt = self[
-                    self._concordances[ctype][
-                        forms[0][0],
-                        concept,
-                        forms[0][2],
-                        forms[0][1],
-                    ][0][0]
-                ]
-                w.writerow([
-                    i,
-                    concept,
-                    sum([f[3] for f in forms]),
-                    ' // '.join(sorted(set([f[2] for f in forms]))),
-                    ' // '.join(sorted(set([format_form(f) for f in forms]))),
+        conc = []
+        for c, refs in getattr(self, ctype).items():
+            if c:
+                igt = self[refs[0][0]]
+                conc.append([
+                    self.clean_lexical_concept(c),
+                    len(refs),
+                    ' // '.join(sorted(set(str(self[ref].gloss) for ref in refs))),
+                    ' // '.join(sorted(set(form(ref) for ref in refs))),
                     igt.phrase_text,
                     igt.gloss_text,
                 ])
+
+        with UnicodeWriter(filename, delimiter='\t') as w:
+            w.writerow(
+                ['ID', 'ENGLISH', 'OCCURRENCE', 'CONCEPT_IN_SOURCE', 'FORMS', 'PHRASE', 'GLOSS'])
+            for i, row in enumerate(sorted(conc, key=lambda x: -x[1]), start=1):
+                w.writerow([i] + row)
         if not filename:
             print(w.read().decode('utf8'))
 
@@ -612,20 +702,24 @@ class Corpus(object):
             ]
         }
         idx = 1
-        for ctype in ['lexicon', 'grammar']:
-            concepts = self.get_concepts(ctype=ctype)
-            concordance = self._concordances[ctype]
-            for concept, entries in concepts.items():
-                for form, lid, cis, freq in entries:
-                    # retrieve the concordance
-                    pidx, sA, sB = concordance[form, concept, cis, lid][0]
-                    txt = self[pidx].phrase
-                    gls = self[pidx].gloss
-                    word, fgls = self[pidx, sA]
+        # Iterate over unique (cleaned concept, form, language, gloss) tuples.
+        i = 0
+        for form, refs in self.form.items():
+            for (lid, gloss), morphrefs in itertools.groupby(
+                sorted(refs, key=lambda r: (self[r[0]].language, str(self[r].gloss))),
+                lambda r: (self[r[0]].language, str(self[r].gloss))
+            ):
+                morphrefs = list(morphrefs)
+                gm = self[morphrefs[0]]
+                gw = self[morphrefs[0][:2]]
+                igt = self[morphrefs[0][0]]
+                i += 1
+                concepts = \
+                    list(itertools.zip_longest(gm.lexical_concepts, [], fillvalue='lexicon')) + \
+                    list(itertools.zip_longest(gm.grammatical_concepts, [], fillvalue='grammar'))
+                for concept, ctype in concepts:
+                    concept = self.clean_lexical_concept(concept)
                     tokens = tokenize(form)
-                    references = ' '.join(
-                        ['{0}:{1}:{2}'.format(a, b, c)
-                         for a, b, c in concordance[form, concept, cis, lid]])
                     # check tokens
                     try:
                         with_lingpy().tokens2class(tokens, 'sca')
@@ -636,26 +730,20 @@ class Corpus(object):
                         D[idx] = [
                             doculect if self.monolingual else lid,
                             concept,
-                            cis,
+                            gloss,
                             ctype,
                             form,
                             tokens,
-                            freq,
-                            word,
-                            fgls,
-                            txt,
-                            gls,
-                            references]
+                            len(morphrefs),
+                            ' '.join(m.form for m in gw),
+                            ' '.join(m.gloss for m in gw),
+                            igt.phrase_text,
+                            igt.gloss_text,
+                            ' '.join('{}:{}:{}'.format(*ref) for ref in morphrefs)]
                         idx += 1
                     else:
                         print('[!] Problem with "{0}" / [{1}] [{2}] / {3} {4} {5}'.format(
-                            concept,
-                            form,
-                            tokens,
-                            pidx,
-                            sA,
-                            sB,
-                        ))
+                            concept, form, tokens, *morphrefs[0]))
         wl = with_lingpy().Wordlist(D)
 
         if lexstat:
@@ -666,7 +754,7 @@ class Corpus(object):
             wl.renumber('cog', ref)
         return wl
 
-    def get_profile(self, clts=None, filename=None):
+    def get_profile(self, clts=None, filename=None) -> segments.Profile:
         """
         Compute an orthography profile with LingPy's function.
 
@@ -676,8 +764,8 @@ class Corpus(object):
         clts = clts.bipa if clts else None
 
         D = {0: ['doculect', 'concept', 'ipa']}
-        for i, key in enumerate(self._concordances.form, start=1):
-            D[i] = ['dummy', key[1], key[0]]
+        for i, key in enumerate(self.form, start=1):
+            D[i] = ['dummy', str(self[self.form[key][0]].gloss), key]
         wordlist = with_lingpy().basic.wordlist.Wordlist(D)
 
         if not filename:
